# Generated by Django 3.1.2 on 2020-10-19 07:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('units', '0002_auto_20201019_0718'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customunit',
            name='symbol',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='Symbol'),
        ),
    ]
