# Generated by Django 3.2 on 2021-04-18 19:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rates', '0003_auto_20210318_0713'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rate',
            name='base_currency',
            field=models.CharField(db_index=True, default='EUR', max_length=3, verbose_name='Currency to convert to'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='currency',
            field=models.CharField(db_index=True, max_length=3, verbose_name='Currency to convert from'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='key',
            field=models.CharField(db_index=True, default=None, max_length=255, null=True, verbose_name='User defined categorization key'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='value',
            field=models.FloatField(default=0, verbose_name='Rate conversion factor'),
        ),
        migrations.AlterField(
            model_name='rate',
            name='value_date',
            field=models.DateField(verbose_name='Date of value'),
        ),
    ]
