"""
Currencies app configuration
"""

from django.apps import AppConfig


class CurrenciesConfig(AppConfig):
    """
    Currencies configuration
    """
    name = "geocurrency.currencies"
    verbose_name = "Currencies"
