"""
Converters app configuration
"""

from django.apps import AppConfig


class ConvertersConfig(AppConfig):
    """
    Converters configuration
    """
    name = "geocurrency.converters"
    verbose_name = "Converters"
