"""
Geocurrency Rates app config
"""

from django.apps import AppConfig


class RatesConfig(AppConfig):
    """
    Rates configuration
    """

    name = "geocurrency.rates"
    verbose_name = "Rates"
